from cag.framework import GenericAnnotator
from cag.graph_elements.base_graph import BaseGraph

from pyArango.graph import EdgeDefinition

import cag.utils as utils
from cag.utils.config import *

from ..test_nodes import *


class AnyAnnotator(GenericAnnotator):
    def __init__(self, conf: Config, params={'mode': 'run-1'}, filter_annotatable=True):
        super().__init__(query=f"""FOR dp IN {AnyGraphCreator._ANY_DATASET_NODE_NAME}
        RETURN dp
        """, params=params, conf=conf, filter_annotatable=filter_annotatable)
    def update_graph(self, timestamp):
        return super().update_graph(timestamp)

class TestAnnotatorBasics:
    def test_simple_fetch(self):
        conf = configuration(graph='SampleGraph', database=whoami())
        gc = AnyGraphCreator("", config=conf)

        conf.arango_db[AnyGraphCreator._ANY_DATASET_NODE_NAME].insert_many([
            {
                '_key': 'test_1',
                'name': 'test_name'
            },
            {
                '_key': 'test_2',
                'name': 'Hello'
            },
            {
                '_key': 'test_3',
                'name': 'World'
            },
            {
                '_key': 'test_4',
                'name': 'ArangoDB'
            }
        ])
        annotator = AnyAnnotator(conf)

        assert len(annotator.fetch_annotator_data()) == 4

        docs = list(annotator.fetch_annotator_data())
        docs[0]['ann'] = base64.b64encode(
            docs[0]['name'].encode("utf-8")).decode('utf-8')
        annotator.complete_annotation(docs[0])

        # after completing one annotation only three docs should remain
        conf = configuration(graph='SampleGraph', database=whoami())
        annotator = AnyAnnotator(conf)
        assert len(annotator.fetch_annotator_data()) == 3

        # however, when changing the params it should have 4 again!
        conf = configuration(graph='SampleGraph', database=whoami())
        annotator = AnyAnnotator(conf, params={'mode': 'run-2'})
        assert len(annotator.fetch_annotator_data()) == 4
        # or when disabling the filter: default OFF!
        conf = configuration(graph='SampleGraph', database=whoami())
        annotator = AnyAnnotator(conf, filter_annotatable=False)
        assert len(annotator.fetch_annotator_data()) == 4



import examples.annotation_example
from cag.utils.config import Config

def main():
    config= Config(
        url="http://127.0.0.1:8529",
        user="root",
        password="root",
        database="_system",
        graph="GenericGraph"
    )

    examples.annotation_example.run_sample(config)



if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
import time
from typing import Callable, ClassVar, Dict, Optional

class TimerError(Exception):
    """A custom exception used to report errors in use of Timer class"""

@dataclass
class Timer:
    timers: ClassVar[Dict[str, float]] = {}
    name: Optional[str] = None
    text: str = "Elapsed time: {:0.4f} seconds"
    logger: Optional[Callable[[str], None]] = print
    _start_time: Optional[float] = field(default=None, init=False, repr=False)

    def __post_init__(self) -> None:
        """Add timer to dict of timers after initialization"""
        if self.name is not None:
            self.timers.setdefault(self.name, 0)

    def start(self) -> None:
        """Start a new timer"""
        if self._start_time is not None:
            raise TimerError(f"Timer is running. Use .stop() to stop it")

        self._start_time = time.perf_counter()

    def stop(self) -> float:
        """Stop the timer, and report the elapsed time"""
        if self._start_time is None:
            raise TimerError(f"Timer is not running. Use .start() to start it")

        # Calculate elapsed time
        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None

        # Report elapsed time
        if self.logger:
            self.logger(self.text.format(elapsed_time))
        if self.name:
            self.timers[self.name] += elapsed_time

        return elapsed_time


from cag.graph_elements.base_graph import BaseGraph

from pyArango.graph import Graph, EdgeDefinition

import cag.utils as utils
from cag.framework import GraphCreatorBase
import datetime
import inspect
class SampleGraph(BaseGraph):
    _edgeDefinitions = [EdgeDefinition('GenericEdge', fromCollections=['GenericNode'], toCollections=['GenericNode'])]
    _orphanedCollections = []
    
def whoami():
    frame = inspect.getouterframes(inspect.currentframe())[1]
    return frame.function


class AnyGraphCreator(GraphCreatorBase):
    _ANY_DATASET_NODE_NAME = "AnyDataset"
    _ANY_EDGE_PUB_CORPUS = "AnyEdgeDSCorpus"
    _name = "Any Graph Creator"
    _description = "Creates a graph based on any corpus"
    _edge_definitions = [
        {
            'relation': _ANY_EDGE_PUB_CORPUS,
            'from_collections': [_ANY_DATASET_NODE_NAME],
            'to_collections': [GraphCreatorBase._CORPUS_NODE_NAME]
        }
    ]

    def __init__(self, corpus_dir, config, initialize=False):
        super().__init__(corpus_dir, config, initialize)

    def init_graph(self):
        corpus = self.create_corpus_node(key="AnyCorpus",
                                           name=AnyGraphCreator._name,
                                           type="journal",
                                           desc=AnyGraphCreator._description,
                                           created_on=datetime.datetime.today())




from .annotator.annotator_base import GenericAnnotator
from .creator.graph_creator_base import GraphCreatorBase
from .analyzer.analyzer_base import AnalyzerBase
__all__ = [GenericAnnotator, GraphCreatorBase, AnalyzerBase]



from cag.framework.annotator import GenericAnnotator
from cag.framework.annotator.pipeline.pipeline_base import Pipeline
from cag.utils.config import *

from examples.graph_creation_example import AnyGraphCreator
from pyArango.query import AQLQuery

from pyArango.collection import Collection

class AnyAnnotator(GenericAnnotator):
    def __init__(self, conf: Config, params={'mode': 'run-1'}, filter_annotatable=True):
        super().__init__(query=f"""FOR dp IN {AnyGraphCreator._ANY_DATASET_NODE_NAME}
        RETURN dp
        """, params=params, conf=conf, filter_annotatable=filter_annotatable, run=True)

    def update_graph(self, timestamp, data: AQLQuery):
        for dp in data:
            pass


## Pipeline for specialized annotator

class SamplePipeline(Pipeline):

     def process_input(self) -> list:
        processed = []
        for txt_node in self.input:
            processed.append((txt_node.text, {"_key": txt_node._key}))

        return processed

     def init_and_run(self):
        # Extend the Default toml if needed, by either defining the path of a dictionary
         # Pipeline.extend_config(extened_dict)

        # Add the pipes defined in the Toml files
         self.add_annotation_pipe(name = "NamedEntityAnnotator", save_output= True, is_spacy=True)
         self.add_annotation_pipe(name = "EmpathAnnotator", save_output= True, is_spacy=True)
         self.add_annotation_pipe(name = "DummyDefaultAnnotator", save_output= False, is_spacy=False)

        # Get your Input in any way you like
         coll: Collection = self.database_config.db["TextNode"]
         docs = coll.fetchAll(limit=50)

        # Set the INPUT - this will automatically call process_input (make sure to implement it
         self.set_input(docs)

        # annotate the input
         self.set_spacy_language_model("en_core_web_md")
         self.annotate()
        # save annotations when "save_output" is set to True
         self.save()

def run_sample(config):
    sample_pipeline = SamplePipeline(config)
    sample_pipeline.init_and_run()



from pyArango.collection import Collection, Field

"""BASE NODES (OOS nodes and Annotations)"""


class GenericNode(Collection):
    _fields = Collection._fields


class GenericOOSNode(Collection):
    _fields = {
        'timestamp': Field()
    }


class AnnotationNode(Collection):
    _fields = {
        'timestamp': Field(),
        'parameters': Field()
    }


"""GENERAL OOS NODES"""


class Corpus(GenericOOSNode):

    _fields = {
        'name': Field(),
        'type': Field(),
        'description': Field(),
        'created_on': Field(),
        **GenericOOSNode._fields
    }


class ImageNode(GenericOOSNode):
    _fields = {
        "url": Field(),
        **GenericOOSNode._fields
    }


class DataNode(GenericOOSNode):
    _fields = {
        "url": Field(),
        **GenericOOSNode._fields
    }


class TextNode(GenericOOSNode):
    _fields = {
        "text": Field(),
        **GenericOOSNode._fields
    }


class Author(GenericOOSNode):
    _fields = {
        "name": Field(),
        **GenericOOSNode._fields
    }



class KeyTerm(GenericOOSNode):
    _fields = {
        "name": Field(),
        **GenericOOSNode._fields
    }


class AbstractNode(GenericOOSNode):
    _fields = {
        "text": Field(),
        **GenericOOSNode._fields
    }


class WebResource(GenericOOSNode):
    _fields = {
        'url': Field(),
        **GenericOOSNode._fields
    }




from cag.utils.config import configuration

from ..test_nodes import *
from ..text_db_fixture import *

class TestGCBasics:
    def test_arango_connection(self):
        config = configuration(database=whoami())
        assert config.arango_db.name == whoami()

    def test_gc_creation(self):
        config = configuration(graph='SampleGraph', database=whoami())
        gc = AnyGraphCreator("", config)
        assert config.arango_db.has_collection(
            AnyGraphCreator._ANY_DATASET_NODE_NAME)
        assert config.arango_db.has_collection(
            AnyGraphCreator._ANY_EDGE_PUB_CORPUS)
        assert not config.arango_db[AnyGraphCreator._CORPUS_NODE_NAME].has(
            'AnyCorpus')

    def test_gc_creation(self):
        config = configuration(graph='SampleGraph', database=whoami())
        gc = AnyGraphCreator("", config, initialize=True)
        assert config.arango_db[AnyGraphCreator._CORPUS_NODE_NAME].has(
            'AnyCorpus')



from arango import ArangoClient

from cag import logger
from cag.view_wrapper.arango_analyzer import ArangoAnalyzer, AnalyzerList
from cag.view_wrapper.link import Link, Field
from cag.view_wrapper.view import View


def run_sample(config):
    db = config.arango_db

    # Create analyzer
    analyzer = ArangoAnalyzer("analyzer_sample")
    analyzer.set_stopwords(language="english", custom_stopwords=['stop', 'word'], include_default=False)
    analyzer.type = ArangoAnalyzer._TYPE_TEXT

    analyzer.create(db)

    # Create Link - a view can hvae 0 to * links
    link = Link(name="TextNode")  # Name of a collection in the database
    linkAnalyzers = AnalyzerList(["identity"])
    link.analyzers = linkAnalyzers

    # A link can have 0..* fields
    field = Field("text", AnalyzerList(
        ["text_en", "invalid_analyzer", "analyzer_sample"]))  # text_en is a predifined analyzer from arango
    logger.info(field.analyzers)
    field.analyzers.filter_invalid_analyzers(db,
                                             verbose=1)  # filters out the analyzer that are not defined in the database
    print("current analyzers after filtering invalid ones: ", field.analyzers)
    ## current analyzers after filtering invalid ones:  AnalyzerList(analyzerList=['text_en', 'analyzer_sample'])

    link.add_field(field)

    ## Show the dict format of all the fields in a link
    print(link.get_fields_dict())
    # {'text': {'analyzers': ['text_en', 'analyzer_sample']}}

    # create view
    view = View('sample_view',
                view_type="arangosearch")
    ## add the link (can have 0 or 1 link)
    view.add_link(link)

    ## can have 0..* primary sort
    view.add_primary_sort("text", asc=False)
    view.add_stored_value(["text", "timestamp"], compression="lz4")

    print("Prints the *view* as a dict:", view.summary())


    ## creates the view in the database
    view.create(db)





from pyArango.document import Document
from cag import logger
from cag.framework.annotator.element.orchestrator import PipeOrchestrator


class DummyPipeOrchestrator(PipeOrchestrator):

    def create_edge(self, _from: Document, _to: Document, **kwargs) -> Document:
        pass


    def create_node(self, **kwargs) -> Document:
        pass

    def save_annotations(self, annotated_text: "[]"):
        pass


def customized_pipe_func(input):
    logger.debug("checking Dummy customized pipe")
    return input


from abc import ABC, abstractmethod
from typing import ClassVar, Dict
from pyArango.document import Document

from cag import logger
from cag.framework.component import Component
from cag.utils import utils
from cag.utils.config import Config


class PipeOrchestrator(ABC, Component):

    ANNOTATION_TYPE : ClassVar = ["text", "url", "image"]
    ANNOTATION_LEVEL: ClassVar = ["node", "corpus"]

    def __init__(self, registered_pipes: Dict, orchestrator_config_id :str  =None, conf: Config = None):
        super().__init__(conf)
        self.name = self.__class__.__name__ if orchestrator_config_id is None else orchestrator_config_id
        self.pipe_config = registered_pipes[self.name]
        self.annotation_level = self.pipe_config["level"]
        self.annotation_type = self.pipe_config["data_type"]

        self.pipe_id_or_func = self.pipe_config["pipe_id_or_func"] if "pipe_id_or_func" in self.pipe_config.keys() else None
        self.pipe_path = self.pipe_config["pipe_path"] if "pipe_path" in self.pipe_config.keys() else None

        self.node_class_path = self.pipe_config ["node_class"]
        self.edge_class_path = self.pipe_config ["edge_class"]
        self.annotated_node = self.pipe_config["annotated_node_name"]


        self.node_name = self.node_class_path.split(".")[-1]
        self.edge_name = self.edge_class_path.split(".")[-1]

        self.load_pipe_component()
        self.init_graph_elts()
        is_valid, errors = self.validate()
        logger.info(f"orchestrator is validated: {is_valid}")
        logger.debug(f"orchestrator is validated: {errors}")


    def init_graph_elts(self):
        self.node_class, annotation_node = utils.get_cls_from_path(self.node_class_path)
        self.edge_class, relation = utils.get_cls_from_path(self.edge_class_path)

        utils.load_module(".".join(self.node_class_path.split(".")[:-1]))
        utils.load_module(".".join(self.edge_class_path.split(".")[:-1]))

        logger.info(f"saving relation {relation} to {annotation_node} and from {self.annotated_node}")
        self.graph.update_graph_structure(relation,
                                          [self.annotated_node],
                                          [annotation_node],
                                          create_collections=True
                                          )

    def load_pipe_component(self):
        module = None
        if self.pipe_path is not None and len(self.pipe_path) > 0:
            logger.debug("loading {}".format(self.pipe_path))
            module = utils.load_module(self.pipe_path)
        return module

    def get_pipe_func(self):
        module = self.load_pipe_component()
        pipe_func = None
        if module is not None and self.pipe_id_or_func is not None:
            pipe_func = getattr(module, self.pipe_id_or_func)
        return pipe_func

    def validate(self):
        error_dict = {}
        if self.annotation_type not in PipeOrchestrator.ANNOTATION_TYPE:
            error_dict["annotation_type"] = "The annotation type should have one of the following values: {} but has the value {}".\
                format(str(PipeOrchestrator.ANNOTATION_TYPE), self.annotation_type)
            logger.error(error_dict["annotation_type"])

        if self.annotation_level not in PipeOrchestrator.ANNOTATION_LEVEL:
            error_dict["annotation_level"] = "The annotation level should have one of the following values: {} " \
                                         "but has the value {}".\
                                          format(str(PipeOrchestrator.ANNOTATION_LEVEL), self.annotation_level)
            logger.error(error_dict["annotation_level"])

        if not self.database.hasCollection(self.annotated_node):
            error_dict["annotated_node"] = "We couldnt find the *annotated_node* {}.".format(self.annotated_node)
            logger.error(error_dict["annotated_node"])

        if not self.database.hasCollection(self.node_name):
            error_dict["node_class"] = "The node was not created, *node_class_path*: {}.".format(self.node_class_path)
            logger.error(error_dict["node_class"])

        if not self.database.hasCollection(self.edge_name):
            error_dict["edge_class"] = "The edge was not created, *edge_class_path*: {}.".format(self.edge_class_path)
            logger.error(error_dict["edge_class"])

        is_valid = not bool(error_dict)
        return is_valid, error_dict


    @abstractmethod
    def create_node(self, **kwargs) -> Document:
        pass

    @abstractmethod
    def create_edge(self, _from: Document, _to: Document, **kwargs) -> Document:
        pass

    @abstractmethod
    def save_annotations(self, annotated_text: "[]"):
        pass





import logging
from dataclasses import dataclass, field
from arango.database import StandardDatabase

from cag import logger
from cag.utils import utils
from arango import AnalyzerGetError
from nltk.corpus import stopwords
from typing import List, ClassVar


@dataclass
class EdgeNGram:
    min: int = 2
    max: int = 4
    preserve_original: bool = True

    def summary(self):
        result = utils.camel_nest_dict(utils.filter_dic(self))
        return result


@dataclass
class ArangoAnalyzer():
    #####################
    ### Instance Vars ###
    #####################
    name: str
    type:str = "identity"
    features: List = field(default_factory=lambda: ["frequency", "norm", "position"])

    ## Properties - loaded based on type
    # TEXT
    locale: str = "en"  # text, stem, norm
    case: str = "lower"  # text,  norm
    stopwords: List = field(default_factory=lambda: [])
    accent: bool = False  # text,  norm
    stemming: bool = True
    edge_ngram: "EdgeNGram | None" = None  #
    # type: delimiter
    delimiter: str = ','

    # type ngram
    min: int = 2
    max: int= 5
    preserve_original: bool = False
    start_marker: str = ""
    end_marker: str = ""
    stem_type: str = field(default_factory=lambda:ArangoAnalyzer._STEM_TYPE_BINARY)


    #################
    ### CONSTANTS ###
    #################

    ######################################
    ### Main types of arango analyzer ###
    #####################################
    _TYPE_IDENTITY: ClassVar = "identity"
    _TYPE_TEXT: ClassVar = "text"
    _TYPE_NGRAM: ClassVar ="ngram"
    _TYPE_STEM: ClassVar = "stem"
    _TYPE_NORM: ClassVar = "norm"
    _TYPE_DELIMITER: ClassVar ="delimiter"

    # For type Stem - there are 2 stem types
    _STEM_TYPE_BINARY: ClassVar = "binary"
    _STEM_TYPE_UTF8: ClassVar = "utf8"

    _MANDATORY_FIELDS: ClassVar = ['name', 'type', 'features']

    FIELDS: ClassVar = {
        _TYPE_IDENTITY : [],
        _TYPE_TEXT: ['locale', 'case', 'stopwords', 'accent', 'stemming', 'edge_ngram'],
        _TYPE_NGRAM: ['min', 'max', 'preserve_original',
                      'start_marker', 'end_marker', 'stem_type'],
        _TYPE_STEM: ['locale'],
        _TYPE_DELIMITER: ['delimiter'],
    }

    #############
    ## Setters ##
    #############
    def set_features(self, frequency= True, norm=True, position = True):
        '''
        Sets which features to return - It is set to have all three features: frequency, norm and position
        :param frequency: bool
        :param norm: bool
        :param position: bool
        :return:
        '''
        features = []
        if frequency: features.append("frequency")
        if norm: features.append("norm")
        if position: features.append("position")
        self.features = features

    def set_edge_ngrams(self, min =2, max = 5, preserve_original = False):
        self.edge_ngram =  EdgeNGram(min =min, max = max, preserve_original = preserve_original)


    def set_stopwords(self, language = "english", custom_stopwords= [], include_default = True):
        logger.info('The default stopword list is loaded from NLTK - if you get an error, run `nltk.download('
                     '"stopwords")`')
        if include_default:
            self.stopwords = stopwords.words(language)

        if len(custom_stopwords) > 0:
            self.stopwords.extend(custom_stopwords)


    #############
    ## Getters ##
    #############
    def get_type_fields(self) -> List:
        '''
        Gets the list of fields needed and set for the Analyzer based on the Type

        :return: list of fields
        '''
        return ArangoAnalyzer.FIELDS[self.type]

    def get_properties(self) -> dict:
        keep = ArangoAnalyzer.FIELDS[self.type]
        result = utils.camel_nest_dict(utils.filter_dic(self, keep))
        return result

    def summary(self) -> dict:
        keep = ArangoAnalyzer._MANDATORY_FIELDS + ArangoAnalyzer.FIELDS[self.type]
        result = utils.camel_nest_dict(utils.filter_dic(self, keep))
        return result


    def create(self, database: StandardDatabase):
        result = {}
        try:
            info = database.analyzer(self.name)
            logger.error("Analyzer {} already exists with the following info: {}".format(self.name, info))
            logger.error("You can delete it first using *database.delete_analyzer({})* and that create".format(self.name))
        except AnalyzerGetError:
            result = database.create_analyzer(self.name,
                                              self.type,
                                              self.get_properties(),
                                              self.features)
            logger.info("Analyzer was created!")
        return result



@dataclass
class AnalyzerList:
    analyzerList: List = field(default_factory=lambda:[])  # ["text_en"]
    database: StandardDatabase = None

    def get_invalid_analyzers(self, database: StandardDatabase, verbose=0):
        invalid_analyzer = []
        for analyzer in self.analyzerList:
            try:
                info = database.analyzer(analyzer)
                if verbose == 1:
                    logger.info(info)
            except AnalyzerGetError:
                invalid_analyzer.append(analyzer)
        return invalid_analyzer

    def filter_invalid_analyzers(self, database: StandardDatabase, verbose:int=0):
        invalid_analyzer = self.get_invalid_analyzers(database)
        if verbose == 1:
            if len(invalid_analyzer) > 0 :
                logger.warning("Filtered out the following invalid analyzers: "+ str(invalid_analyzer))
        self.analyzerList = [a for a in self.analyzerList if a not in invalid_analyzer]


    def add_analyzer(self, analyzer_name:str, force_add = False):
        '''
        Adds the analyzer name to a list. If force_add =True then it adds is even if this analyzer does not exist in the database
        '''
        try:
            if self.database is not None:
                _ = self.database.analyzer(analyzer_name)
            else:
                print("couldnt validate analyzer because database connection is not set.")
            self.analyzerList.append(analyzer_name)
        except AnalyzerGetError:
            print("Invalid Analyzer")
            if force_add:
                self.analyzerList.append(analyzer_name)




import os
import platform
import sys
import zipfile
from pathlib import Path
from subprocess import PIPE, run

import typer
from slugify import slugify
from rich import print

app = typer.Typer()


def setup_project_name():
    return typer.prompt("Enter desired project name", type=str)


def setup_project_dir(project_name: str):
    cwd = os.getcwd()
    project_dir = Path(cwd).joinpath(slugify(project_name))
    if project_dir.exists():
        print(f"Aborting. {project_dir} already existent!")
        raise typer.Abort()
    if not typer.confirm(
            f"Set {project_dir} as the project directory. Is this correct?", default=True
    ):
        print("Aborting...")
        raise typer.Abort()
    print(f"Creating {project_name}")
    project_dir.mkdir()
    return project_dir


@app.command()
def show_info():
    """
    Use this info when opening issues!
    """
    command = ['pip', 'show', 'cag']
    result = run(command, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    print(f"Package info\n{result.stdout}")
    print(10*"=")
    print(f"Python version {platform.python_version()}")
    print(10*"=")
    print(f"OS: {platform.platform(aliased=True)}")


@app.command()
def start_project():
    """
    Creates project scaffold to get you started
    """
    name = setup_project_name()
    project_dir = setup_project_dir(name)
    path = Path(os.path.dirname(__file__))  # get path where the module is
    with zipfile.ZipFile(f"{path.joinpath('cag_sample_project.zip')}", "r") as zip_ref:
        zip_ref.extractall(project_dir)
    print(f"[green] Creating scaffold in {project_dir}[/green]")


if __name__ == "__main__":
    app()



from pyArango.graph import Graph, EdgeDefinition

import json

from pyArango.theExceptions import CreationError

from cag import logger


class BaseGraph(Graph):
    _edgeDefinitions = [EdgeDefinition('GenericEdge', fromCollections=['GenericNode'], toCollections=['GenericNode'])]
    _orphanedCollections = []

    def _check_and_update_collections(self, collections):
        """Creates new collections from the given collections list if not yet existant."""
        for col in collections:
            if not self.database.hasCollection(col):
                self.database.createCollection(col)

    def update_graph_structure(self, relation, from_collections, to_collections, create_collections=True, waitForSync = False):
        """
        Adds an edge definition to the graph if it not already exists. Otherwhise the existing edge definition becomes updated if necessary based on the given collections.
        In case of an update, this method never deletes collections from an edge definition but creates a union of the existing and provided collections.

        Parameters:
            relation (required string): Name of the edge collection defining the relation.
            from_collections (required [string]): List of names of node collections appearing as origin of edges.
            to_collections (required [string]): List of names of node collections appearing as destination of edges.
            create_collections (optional boolean, default: true): If true collections will be created if they not exist.
            waitForSync (optional boolean, default: False): See ArangoDBs doc for possible col arguments.
        Returns:
            Boolean value indicating if the edge definitions of this graph have been updated.
        """

        if relation in self.definitions.keys():

            existings_froms = set(self.definitions[relation].fromCollections)
            existing_tos = set(self.definitions[relation].toCollections)

            proposed_froms = set(from_collections)
            proposed_tos = set(to_collections)

            new_froms = proposed_froms - existings_froms
            new_tos = proposed_tos - existing_tos

            if create_collections and len(new_froms) > 0:
                self._check_and_update_collections(new_froms)


            if create_collections and (len(new_tos) > 0):
                self._check_and_update_collections(new_tos)

            if (len(new_froms ) > 0) or (len(new_tos) > 0):
                 url = '%s/edge/%s' % (self.getURL(), relation)
                 from_collections = list(existings_froms | new_froms)
                 to_collections = list(existing_tos | new_tos)
            else:
                logger.debug(f"edge  found {relation} but has no changes")
                return False

            logger.debug(f"edge  found {relation}, updating it")
            r = self.connection.session.put(url, data=json.dumps(
                {
                    'collection': relation,
                    'from': from_collections,
                    'to': to_collections
                }, default=str), params={'waitForSync': waitForSync}
                                            )
        else:
            logger.debug(f"edge not found {relation}, adding it")
            url = '%s/edge' % self.getURL()
            r = self.connection.session.post(url, data=json.dumps(
                {
                    'collection': relation,
                    'from': from_collections,
                    'to': to_collections
                },
                default=str), params={'waitForSync': waitForSync}
                                             )

        data = r.json()
        if r.status_code == 201 or r.status_code == 202:
            self.definitions[relation] = EdgeDefinition(relation, fromCollections = from_collections, toCollections = to_collections)
            return True

        raise CreationError("Unable to modify edge definitions., %s" % data["errorMessage"], data)


from cag.framework import AnalyzerBase
from cag.graph_elements.base_graph import BaseGraph

from pyArango.graph import EdgeDefinition

import cag.utils as utils
from cag.utils.config import *

from ..test_nodes import *
import networkx as nx


class AnylzerGC(GraphCreatorBase):
    _ANY_DATASET_NODE_NAME = "AnyDataset"
    _ANY_EDGE_PUB_CORPUS = "AnyEdgeDSCorpus"
    _name = "Any Graph Creator"
    _description = "Creates a graph based on any corpus"
    _edge_definitions = [
        {
            'relation': _ANY_EDGE_PUB_CORPUS,
            'from_collections': [_ANY_DATASET_NODE_NAME],
            'to_collections': [GraphCreatorBase._CORPUS_NODE_NAME, _ANY_DATASET_NODE_NAME]
        }
    ]

    def __init__(self, corpus_dir, config, initialize=True):
        super().__init__(corpus_dir, config, initialize)

    def init_graph(self):
        corpus = self.create_corpus_node(key="analyzer_corpus",
                                           name=AnyGraphCreator._name,
                                           type="journal",
                                           desc=AnyGraphCreator._description,
                                           created_on=datetime.datetime.today())
        prev_vert = corpus
        for i in range(10):
            doc = {
                '_key': f"test_{i}",
                'text': f"TEST TEXT {i}"
            }
            vert = self.upsert_node(AnylzerGC._ANY_DATASET_NODE_NAME, doc)
            self.upsert_edge(AnylzerGC._ANY_EDGE_PUB_CORPUS, prev_vert, vert)
            prev_vert = vert


class AnyAnalyzer(AnalyzerBase):
    def __init__(self, conf: Config):
        super().__init__(config=conf, run=True, query="""
         FOR path IN 1..10 ANY K_PATHS 'Corpus/analyzer_corpus' TO 'AnyDataset/test_9' AnyEdgeDSCorpus
                RETURN path
        """)

    def run(self, data):
        data = list(data)
        self.networkx = self.create_networkx(data, weight_edges=True)


class TestAnnotatorBasics:
    def test_create_nx(self):
        
        conf = configuration(graph='SampleGraph', database=whoami())
        gc = AnylzerGC("", config=conf)
        # act
        an = AnyAnalyzer(conf)
        g = an.networkx
        # assert that nx build correctly
        avg_l = nx.average_shortest_path_length(g)
        assert avg_l==4.0



import setuptools
from setuptools import setup
__version__  ="0.5.5"
URL = 'https://github.com/DLR-SC/corpus-annotation-graph-builder'
setup(
    name='cag',
    version=__version__,
    description='This is a general framework to create arango db graphs and annotate them.',
    url=URL,
    author='DLR',
    author_email='roxanne.elbaff@dlr.de, tobias.hecking@dlr.de',
    license='MIT',
    download_url=f'{URL}/archive/{__version__}.tar.gz',
    packages=setuptools.find_packages(),
    python_requires='>=3.7',
    keywords=['graph', 'architectural framework', 'graph creator', 'graph annotator'],
    install_requires=['dataclasses>=0.6',
                    'spacy>=3.4.1',
                    'spacy_arguing_lexicon>=0.0.3',
                    'empath>=0.89',
                    'pytest>=7.1.2',
                    'networkx>=2.8.5',
                    'nltk>=3.4.5',
                    'pyvis>=0.2.1',
                    'tqdm>=4.43.0',
                    'python-arango>=7.4.1',
                    #'arango==0.2.1',
                    'pyArango>=2.0.1',
                    'tomli>=2.0.1'
                      ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
    ],
    entry_points={
        'console_scripts': ['cag=cag.cli:app'],
    },
    zip_safe=False
)



from queue import Queue
from abc import ABC, abstractmethod
from typing import ClassVar

import dataclasses
from pyArango.collection import Collection
from spacy import Language

from cag import logger
from cag.framework.annotator import registered_pipes
from cag.framework.annotator.element.orchestrator import PipeOrchestrator
from cag.utils import utils
import spacy
from tqdm import tqdm
from cag.utils.config import Config

@dataclasses.dataclass
class Pipe:
    name: str
    is_spacy:bool = False
    save_output:bool = True
    pipe_id_or_func:str = None # set internally from toml




class Pipeline(ABC):
    REGISTERED_PIPE_CONFIGS: ClassVar = registered_pipes._dict

    def _load_registered_pipes(self, load_default_pipe_configs:bool = True,
                 extended_pipe_configs:dict = None):
        if load_default_pipe_configs:
            _copy = Pipeline.REGISTERED_PIPE_CONFIGS.copy()
            if  extended_pipe_configs is not None:
                _copy.update(extended_pipe_configs)
            registered_pipes = _copy
        return registered_pipes
    def __init__(self,
                 database_config: Config,
                 input: "[Collection]" = None,
                 load_default_pipe_configs = True,
                 extended_pipe_configs:dict = None
                 ):
        """
        The Pipeline class is responsible for taking a set of nodes as input, annotating them and saving them into
        the database.

        """

        self.registered_pipes = self._load_registered_pipes(load_default_pipe_configs, extended_pipe_configs)

        self.pipe_instance_dict = {}
        self.pipeline:[Pipe] = []


        self.database_config = database_config

        self.processed_input = None
        self.set_input(input)

        self.annotated_artifacts = None

        self.set_spacy_language_model()

    ############################################
    #####      ABSTRACT METHODS            #####
    ############################################
    @abstractmethod
    def process_input(self) -> object:
        pass

    @abstractmethod
    def init_and_run(self) -> list:
        pass
    ############################################
    #####      MAIN. ANNOTATE and SAVE     #####
    ############################################

    def annotate(self):
        pipe_stacks = self.get_pipe_stacks()
        input = self.processed_input
        out = None
        while not pipe_stacks.empty():
            pipe_stack = pipe_stacks.get()
            if pipe_stack['stack_type'] == 'spacy':
                nlp = self.init_spacy_nlp(pipe_stack['stack'])
                out = list(nlp.pipe(input, as_tuples=True))
                input = out
            else:
                while pipe_stack['stack']:
                    current_pipe: Pipe = pipe_stack['stack'].pop(0)
                    pipe_instance :PipeOrchestrator = self.pipe_instance_dict[current_pipe.name]
                    pipe_func = pipe_instance.get_pipe_func()
                    if pipe_func is None:
                        logger.error("The pipe is not a spacy pipe. Make sure to define the pipe_path and pipe code and "
                                     "provide the implementation within the pipe path and a function with the name equivalent to the "
                                     "'pipe' name")
                    else:
                        out = pipe_func(input)
                        input = out
        self.annotated_artifacts = out


    def save(self):
        if self.annotated_artifacts is None:
            logger.error("Nothing to save, call annotate before saving")
            pass
        for pipe in tqdm(self.pipeline):
            if pipe.save_output:
                logger.info("saving annotations of {}".format(pipe))
                self.pipe_instance_dict[pipe.name].save_annotations(self.annotated_artifacts)


    ############################################
    #####           SETTERS                #####
    ############################################
    def reset_input_output(self):
        self.set_input(None)
        self.annotated_artifacts = None

    def set_input(self, nodes: [Collection]):
        self.input = nodes
        if self.input is not None:
            self.processed_input = self.process_input()
        else:
            self.processed_input = None

    def set_spacy_language_model(self, language_package = "en_core_web_sm"):
        self.spacy_language_model = language_package
        self.load_spacy_model()
        # English pipeline optimized for CPU. Components: tok2vec, tagger, parser, senter, ner, attribute_ruler, lemmatizer:
        #   en_core_web_sm (12MB)
        #   en_core_web_md (31MB)
        #   en_core_web_lg (382MB)
        # English transformer pipeline (roberta-base). Components: transformer, tagger, parser, ner, attribute_ruler, lemmatizer.
        #   en_core_web_trf (438)

    def add_annotation_pipe(self, pipe:Pipe=None,
                            name:str="", save_output: bool = False, is_spacy:bool=False):
        """
            The add_annotation_pipe adds a pipe to the pipeline. the pipes are first in first out (FIFO).
            The corrisponding Annotator class (given from the toml)  to this pipe is initiated and saved, to be used
            for annotating and saving to the database.
            *annotator_instances*
            *pipe_id_or_func*
            *pipeline*
        """
        pipe = pipe if pipe is not None else Pipe(name, save_output, is_spacy)

        logger.info(f"adding pipe with name {pipe.name}")
        cls, _ = utils.get_cls_from_path(self.registered_pipes[pipe.name][registered_pipes.PipeConfigKeys._orchestrator_class])
        instance = cls(self.registered_pipes, orchestrator_config_id= pipe.name, conf=self.database_config)
        self.pipe_instance_dict[pipe.name] = instance

        logger.info(f"adding pipe with code {instance.pipe_id_or_func}")
        pipe.pipe_id_or_func = instance.pipe_id_or_func

        self.pipeline.append(pipe)

    ############################################
    #####       HELPER METHODS             #####
    ############################################
    def get_pipe_stacks(self):
        logger.info("Defining pipe default and spacy stacks")
        pipe_queue: [Pipe] = self.pipeline.copy()
        pipe_stack: Queue = Queue()  # {"spacy": ["ner", "bla", "bloo], "default": ["da", "doo", "di"}
        task_collection = []
        previous_task_is_spacy = False
        while pipe_queue:
            current: Pipe = pipe_queue.pop(0) # Fifo
            current_is_spacy = current.is_spacy
            if (previous_task_is_spacy == current_is_spacy) or len(task_collection) == 0:
                task_collection.append(current)
            else:
                stack_type = "spacy" if previous_task_is_spacy else "default"
                pipe_stack.put({'stack_type': stack_type, 'stack': task_collection})
                task_collection = [current]
            previous_task_is_spacy = current_is_spacy

        if len(task_collection) > 0:
            stack_type = "spacy" if previous_task_is_spacy else "default"
            pipe_stack.put({'stack_type': stack_type, 'stack': task_collection})


        return pipe_stack

    def init_spacy_nlp(self, subpipeline) -> Language:
        nlp = spacy.load(self.spacy_language_model, disable=["ner"])
        for pipe in subpipeline:
            if not nlp.has_pipe(pipe.pipe_id_or_func):
                if pipe.pipe_id_or_func in nlp.disabled:
                    nlp.enable_pipe(pipe.pipe_id_or_func)
                else:
                    nlp.add_pipe(pipe.pipe_id_or_func)#, **kargs)
            else:
                logger.debug("pipe already in pipeline")
        return nlp

    def load_spacy_model(self):
        if not spacy.util.is_package(self.spacy_language_model):
            spacy.cli.download(self.spacy_language_model)





from abc import ABC, abstractmethod
from dataclasses import dataclass
from pyvis.network import Network
import networkx as nx
from cag.utils.config import Config
from cag.framework.component import Component
from enum import Enum
from pyArango.query import AQLQuery


class AnalyzerMode(Enum):
    NOTEBOOK = 1
    HTML = 2
    MATPLOTLIB = 3


@dataclass
class Doc:
    _id: str
    _key: str


@dataclass
class EdgeDoc(Doc):
    _from: str
    _to: str
    _id: str
    _key: str


@dataclass
class PathElement:
    vertices: 'list[Doc]'
    edges: 'list[EdgeDoc]'


class AnalyzerBase(ABC, Component):
    def __init__(self, config: Config, mode: AnalyzerMode = AnalyzerMode.NOTEBOOK, run=False, query: "None|str" = None, params={'rawResults': True}) -> None:
        super().__init__(config)
        self.mode = mode
        self.data = []
        self.query = query
        if self.query:
            self.data = self.database.AQLQuery(self.query, **params)
        if run:
            self.run(self.data)

    @abstractmethod
    def run(data):
        pass

    def create_networkx(self, graph_data: 'list[PathElement]', weight_edges=False) -> nx.Graph:
        G = nx.Graph()
        for p in graph_data:
            for v in p['vertices']:
                G.add_node(v['_id'], **v)
            for e in p['edges']:
                f_t = (e['_from'], e['_to'],)
                if weight_edges:
                    G.add_edge(*f_t, **e)
                else:
                    if G.has_edge(*f_t):
                        G.get_edge_data(*f_t)['weight'] += 1
                    else:
                        G.add_edge(*f_t, weight=1, **e)
        return G

    def visualize_graph(self, graph_data: 'list[PathElement]', weight_edges=False) -> Network:
        G = self.create_networkx(graph_data, weight_edges)
        g = Network(notebook=(self.mode == AnalyzerMode.NOTEBOOK))
        g.from_nx(G)
        return g



import dataclasses
from typing import ClassVar

_dict = {}

########## Dictionary Keys ##########
@dataclasses.dataclass
class PipeConfigKeys:
    _orchestrator_class : ClassVar = "orchestrator_class"
    _pipe_id_or_func : ClassVar = "pipe_id_or_func"
    _pipe_path : ClassVar = "pipe_path"  # leave empty if the pipe is a spacy native pipe
    _level : ClassVar = "level"  # node or set
    _data_type : ClassVar = "data_type"  # for now we support text - later url, image
    _annotated_column : ClassVar = "annotated_column"
    _input_id : ClassVar = "input_id"
    _node_class : ClassVar = "node_class"
    _edge_class : ClassVar ="edge_class"
    _annotated_node_name : ClassVar = "annotated_node_name"



_dict["NamedEntityPipeOrchestrator"] = {
    PipeConfigKeys._orchestrator_class : "cag.framework.annotator.instance.ner_orchestrator.NamedEntityPipeOrchestrator",
    PipeConfigKeys._pipe_id_or_func : "ner", # id in case of space, function name otherwise
    PipeConfigKeys._pipe_path : "",  # leave empty if the pipe is a spacy native pipe, otherwise provide the path of where the pipe_id_or_func exists
    PipeConfigKeys._level : "node",  # node or set
    PipeConfigKeys._data_type : "text",  # for now we support text - later url, image
    PipeConfigKeys._annotated_node_name : "TextNode",
    PipeConfigKeys._node_class : "cag.framework.annotator.element.graph_elements.NamedEntityNode",
    PipeConfigKeys._edge_class : "cag.framework.annotator.element.graph_elements.HasAnnotation"
}


_dict["EmpathPipeOrchestrator"] = {
    PipeConfigKeys._orchestrator_class : "cag.framework.annotator.instance.empath_orchestrator.EmpathPipeOrchestrator",
    PipeConfigKeys._pipe_id_or_func : "empath_component", # id in case of space, function name otherwise
    PipeConfigKeys._pipe_path : "cag.framework.annotator.pipe.linguistic.style_features",  # leave empty if the pipe is a spacy native pipe, otherwise provide the path of where the pipe_id_or_func exists
    PipeConfigKeys._level : "node",  # node or set
    PipeConfigKeys._data_type : "text",  # for now we support text - later url, image
    PipeConfigKeys._annotated_node_name : "TextNode",
    PipeConfigKeys._node_class : "cag.framework.annotator.element.graph_elements.EmpathNode",
    PipeConfigKeys._edge_class : "cag.framework.annotator.element.graph_elements.HasAnnotation"
}

_dict["DummyPipeOrchestrator"] = {
    PipeConfigKeys._orchestrator_class : "cag.framework.annotator.instance.dummy.DummyPipeOrchestrator",
    PipeConfigKeys._pipe_id_or_func : "customized_pipe_func", # id in case of space, function name otherwise
    PipeConfigKeys._pipe_path : "cag.framework.annotator.instance.dummy",  # leave empty if the pipe is a spacy native pipe, otherwise provide the path of where the pipe_id_or_func exists
    PipeConfigKeys._level : "node",  # node or set
    PipeConfigKeys._data_type : "text",  # for now we support text - later url, image
    PipeConfigKeys._annotated_node_name : "TextNode",
    PipeConfigKeys._node_class : "cag.graph_elements.nodes.GenericNode",
    PipeConfigKeys._edge_class : "cag.framework.annotator.element.graph_elements.HasAnnotation"
}


import logging

from cag.utils import utils
from cag.view_wrapper.arango_analyzer import AnalyzerList


from dataclasses import field, dataclass
from typing import List

@dataclass
class Field():
    field_name: str
    analyzers:AnalyzerList = field(default_factory=lambda:AnalyzerList())

    def summary(self):
        dict_ = {}
        dict_[self.field_name] = {"analyzers": self.analyzers.analyzerList}
        return dict_

@dataclass
class Link():
    '''
    Check attributes documentations here https://www.arangodb.com/docs/stable/arangosearch-views.html
    '''
    name: str
    include_all_fields: bool = False
    store_values:str =  None
    track_list_positions: bool = False
    in_background: bool = False
    fields: List[Field] = field(default_factory =  lambda:[])
    analyzers:AnalyzerList = field(default_factory=lambda:AnalyzerList(["identity"]))

    def add_field(self, field:Field):
        self.fields.append(field)

    def get_fields_dict(self):
        dict_ = {}
        for f in self.fields:
            dict_.update(f.summary())
        return dict_

    def summary(self):
        '''
        Return Link as dictionary to be used in arango
        :return:
        '''
        keep = ['include_all_fields', 'store_values', 'track_list_positions', 'in_background']
        dict_  =  {}
        dict_[self.name] = {}
        dict_[self.name]["analyzers"] = self.analyzers.analyzerList
        dict_[self.name]["fields"] = self.get_fields_dict()
        dict_[self.name].update(utils.camel_nest_dict(utils.filter_dic(self, keep)))

        return dict_





